
-- module GenEditorInterfaces where

-- library side for command line processing, command description,
-- and editor interface generation for Vim/GVim (Emacs to follow)

import Char(toUpper)
import System
import IO(hPutStrLn,stderr)
import Maybe

import LocalSettings

data    Comment       = Comment String
data    FileNamePar f = FileNamePar Bool f
data    PathPar f     = PathPar String String f
newtype PositionPar f = PositionPar f
newtype RegionPar f   = RegionPar f
data    NamePar f     = NamePar String f
data    OptionPar f   = OptionPar Bool String f

comment c       = Comment c
fileNamePar     = FileNamePar True
fileNameBasePar = FileNamePar False
pathPar prompt  = PathPar prompt ""
pathDefaultPar  = PathPar 
positionPar     = PositionPar
regionPar       = RegionPar
namePar         = NamePar
optionPar       = OptionPar

data Cmd f = Cmd{ cmd         :: f -> [String] -> IO ()
                , description :: String
                , vim         :: String
                , emacs       :: String
                }

data Entry f = Entry String String (Cmd f)
             | Menu String [Entry f]

editParameters entry cname c ps = Entry entry cname
                                        Cmd{cmd=const $ cmd x c
                                           ,description=description x
                                           ,vim=vim x
                                           ,emacs=emacs x
                                           }
  where
    x = editPars (cname,"","","","","",0) ps

errorExit msg = hPutStrLn stderr msg >> exitFailure

class EditPars p f where
  editPars :: (String,String,String,String,String,String,Int) -> p -> Cmd f

instance EditPars Comment (IO ())
  where
    editPars (cname,vs,vl,es,ep,ei,n) (Comment comment) = 
      let it c [] = c
          it c as = errorExit $ "no further parameters expected: "
                             ++ show as
      in Cmd{cmd         = it
            ,description = " "++comment
            ,vim         = unlines
                         [""
                         ,"function! Haskell_refac_"++cname++"()"
                         ,vl
                         --,"\tlet @r=\"\""
                         --,"\tredir @r"
                         -- ,"\tsilent echo system(g:haskell_refac_refactorer_client"
                         --                ++".\" "++cname++"\""++vs++")"
                         ,"call Haskell_refac_client_cmd(\" "++cname++"\""++vs++")"
                         --,"\tredir END"
                         --,"\tif g:haskell_refac_showWindow"
                         --,"\t  call Haskell_refac_msg(@r)"
                         --,"\tendif"
                         ,"endfunction"
                         ]
            ,emacs       = unlines
                     [""
                     ,"(defun haskell-refac-"++cname++" ("++ep++")"
                     ,"  \""++cname++"\""
                     ,"  (interactive \""++ei++"\")"
                     ,"  (let (modified)"
                     ,"     (dolist (b (buffer-list) modified)"
                     ,"       (let* ((n (buffer-name b)) (n1 (substring n 0 1)))"
                     ,"        (if (and (not (or (string= \" \" n1) (string= \"*\" n1))) (buffer-modified-p b))"
                     ,"          (setq modified (cons (buffer-name b) modified)))))"
                     ,"     (if modified (message-box (format \"there are modified buffers: %s\" modified))"
--                     ,"  (apply 'call-process haskell-refac-refactorer "
--                     ,"\t\tnil \"refac\" t"
                     ,"       (process-send-string \"haskell-refac-process\" "
                     ,"         (concat"
                     ,"         "++(unwords $ map show refactor_args)
                     ,"         \""++cname++" \" (mapconcat 'identity (append "++es++") \" \")"
                     ,"         \"\\n\"))"
                     ,"     )"
                     ,"  ) )"
                     ]
            }

instance EditPars ps fps
      => EditPars (OptionPar ps) (String->fps)
  where
    editPars (cname,vs,vl,es,ep,ei,n) (OptionPar isList option ps) = 
      let c = editPars (cname
                       ,vs++".\" \". g:haskell_refac_"++option
                       ,vl
                       ,es++"\n\t\t"++optionString isList "haskell-refac-"
                       ,ep
                       ,ei
                       ,n+1) ps
          it f (a:as) = cmd c (f a) as
          it f []     = errorExit $ "Expected: "++desc
          desc = " <option ("++option++")>"++description c
          optionString False prefix = "(list "++prefix++option++")"
          optionString True  prefix = prefix++option
      in Cmd{cmd         = it
            ,description = desc
            ,vim         = vim c
            ,emacs       = emacs c
            }

instance EditPars ps fps
      => EditPars (NamePar ps) (String->fps)
  where
    editPars (cname,vs,vl,es,ep,ei,n) (NamePar prompt ps) = 
      let c = editPars (cname
                       ,vs++".\" \"."++name
                       ,vl++"\tlet "++name++" = Myinput("++show prompt++")\n"
                       ,es++"\n\t\t(list "++name++")"
                       ,ep++" "++name
                       ,ei++"s"++prompt++"\\n"
                       ,n+1) ps
          it f (a:as) = cmd c (f a) as
          it f []     = errorExit $ "Expected: "++desc
          name = "name"++show n
          desc = " <name ("++prompt++")>"++description c
      in Cmd{cmd         = it
            ,description = desc
            ,vim         = vim c
            ,emacs       = emacs c
            }

instance EditPars ps fps
      => EditPars (PathPar ps) (String->fps)
  where
    editPars (cname,vs,vl,es,ep,ei,n) (PathPar prompt def ps) = 
      let c = editPars (cname
                       ,vs++".\" \"."++name
                       ,vl++"\tlet "++name++" = Myinput("++show prompt++","++show def++")\n"
                       ,es++"\n\t\t"++"(if (eq (length "++name++") 0)"
                          ++"\n\t\t\t(list "++show def++")"
                          ++"\n\t\t\t(list "++name++"))"
                       ,ep++" "++name
                       ,ei++"s"++prompt++"["++(showNoQuotes def)++"]"++"\\n"
                       ,n+1) ps
          it f (a:as) = cmd c (f a) as
          it f []     = errorExit $ "Expected: "++desc
          name = "pathname"++show n
          desc = " <pathname ("++prompt++"["++def++"])>"++description c
      in Cmd{cmd         = it
            ,description = desc
            ,vim         = vim c
            ,emacs       = emacs c
            }

instance EditPars ps fps
      => EditPars (FileNamePar ps) (String->fps)
  where
    editPars (cname,vs,vl,es,ep,ei,n) (FileNamePar keepExt ps) = 
      let c = if keepExt
              then editPars  (cname
                             ,vs++".\" \"."++filename
                             ,vl++"\tlet "++filename++" = expand(\"%:t\")\n"
                             ,es++"\n\t\t(list (file-name-nondirectory (if (buffer-file-name) (buffer-file-name) \"<missing filename>\")))"
                             ,ep
                             ,ei
                             ,n+1) ps
              else editPars  (cname
                             ,vs++".\" \"."++filename
                             ,vl++"\tlet "++filename++" = expand(\"%:t:r\")\n"
                             ,es++"\n\t\t(list (file-name-sans-extension (file-name-nondirectory (if (buffer-file-name) (buffer-file-name) \"<missing filename>\"))))"
                             ,ep
                             ,ei
                             ,n+1) ps
          it f (a:as) = cmd c (f a) as 
          it f []     = errorExit $ "Expected: "++desc
          modifier = if keepExt then "" else "Base"
          filename = "file"++modifier++"Name"++show n
          desc     = " <file"++modifier++"Name>"++description c
      in Cmd{cmd         = it
            ,description = desc
            ,vim         = vim c
            ,emacs       = emacs c
            }

instance EditPars ps fps
      => EditPars (PositionPar ps) (Int->Int->fps)
  where
    editPars (cname,vs,vl,es,ep,ei,n) (PositionPar ps) = 
      let c = editPars (cname
                       ,vs++".\" \"."++line++".\" \"."++col
                       ,vl++"\tlet "++line++" = line(\".\")\n"++"\tlet "++col++" = col(\".\")\n"
                       ,es++"\n\t\t(list (number-to-string (line-no)) (number-to-string (+ 1 (current-column))))"
                       ,ep
                       ,ei
                       ,n+1) ps
          -- need to catch read errors here, for better message
          it f (ls:(cs:as)) = cmd c (f (read ls) (read cs)) as
          it f as           = errorExit $ "Expected: "++desc
                                       ++ "\nFound: "++show as
          line = "line"++show n
          col  = "column"++show n
          desc = " <line> <column>"++description c
      in Cmd{cmd         = it
            ,description = desc
            ,vim         = vim c
            ,emacs       = emacs c
            }

instance EditPars ps fps
      => EditPars (RegionPar ps) (Int->Int->Int->Int->fps)
  where
    editPars (cname,vs,vl,es,ep,ei,n) (RegionPar ps) = 
      let c = editPars (cname
                       ,vs++".\" \"."++start++".\" \"."++end
                       ,vl++"\tlet "++start++" = line(\"'<\").\" \".col(\"'<\")"
                          ++"\n\tlet "++end++" = line(\"'>\").\" \".col(\"'>\")"
                       ,es++"\n\t\t(list (number-to-string (line-no-pos "++start++"))"
                          ++"\n\t\t(number-to-string (+ 1 (current-column-pos "++start++")))"
                          ++"\n\t\t(number-to-string (line-no-pos "++end++"))"
                          ++"\n\t\t(number-to-string (+ 1 (current-column-pos "++end++"))))"
                       ,ep++" "++start++" "++end
                       ,ei++"r"
                       ,n+1) ps
          -- need to catch read errors here, for better message
          it f (ls:(cs:(le:(ce:as)))) = 
            cmd c (f (read ls) (read cs) (read le) (read ce)) as
          it f as = errorExit $ "Expected: "++desc
                             ++ "\nFound: "++show as
          start = "start"++show n
          end   = "end"++show n
          desc  = " <line> <column> <line> <column>"++description c
      in Cmd{cmd         = it
            ,description = desc
            ,vim         = vim c
            ,emacs       = emacs c
            }


gen_vim_interface = do
  putStrLn $ unlines ["\" Vim/GVim interface to Haskell refactorer"
                     ,"\" (generated automatically - manual edits may be lost)\n"
                     ,"\" use :source <file> to load this into vim/gvim\n"
                     ,""
                     ,"let g:haskell_refac_version = \""++hare_version++"\""
                     ,""
                     ,"if !exists(\"g:haskell_refac_refactorer\")"
                     ,"\tlet g:haskell_refac_refactorer = \""++refactorer++"\""
                     ,"endif"
                     ,""
                     ,"if !exists(\"g:haskell_refac_refactorer_client\")"
                     ,"\tlet g:haskell_refac_refactorer_client = \""++refactorer_client++"\""
                     ,"endif"
                     ,""
                     ,"if !exists(\"g:haskell_refac_chasePaths\")"
                     ,"\tlet g:haskell_refac_chasePaths = \". "++showNoQuotes preludePath++"\""
                     ,"endif"
                     ,""
                     ,"if !exists(\"g:haskell_refac_showWindow\")"
                     ,"\tlet g:haskell_refac_showWindow = 1"
                     ,"endif"
                     ,""
                     ,"function! Myinput(p)"
                     ,"     if version >=602"
                     ,"        call inputsave()"
                     ,"     endif"
                     ,"        let x=input(a:p)"
                     ,"     if version >=602"
                     ,"        call inputrestore()"
                     ,"     endif"
                     ,"        return x"
                     ,"endfunction"
                     ,""
                     ]
  mapM_ gen_function cmds
  mapM_ gen_command cmds
  putStrLn $ unlines ["\n\" --------------------------------------------------- menu"
                     ,"if has(\"menu\")"
                     ,"\tif exists(\"b:haskell_refac_menu\")"
                     ,"\t\taunmenu Haskell(Refac)"
                     ,"\tendif\n"
                     ,"\t200amenu Haskell(Refac).-Change-      :"
                     ]
  mapM_ (gen_menu_entry "") cmds
  putStrLn $ unlines ["\tamenu Haskell(Refac).-Custom- :"
                     ,"\tamenu Haskell(Refac).version :echo g:haskell_refac_version<cr>"
                     ,"\tamenu Haskell(Refac).customize :call Customize()<cr>"
                     ,"\ttmenu Haskell(Refac).customize customize paths and options"
                     ,"\tamenu Haskell(Refac).start :RefacStart<cr>"
                     ,"\ttmenu Haskell(Refac).start start refactorer"
                     ,"\tamenu Haskell(Refac).stop :RefacStop<cr>"
                     ,"\ttmenu Haskell(Refac).stop stop refactorer"
                     ,""
                     ,"\tamenu disable Haskell(Refac).*"
                     ,"\tamenu enable Haskell(Refac).version"
                     ,"\tamenu enable Haskell(Refac).customize"
                     ,"\tamenu enable Haskell(Refac).start"
                     ,""
                     ,"\n\tlet b:haskell_refac_menu = \"yes\""
                     ,"endif"
                     ,""
                     ,"function! Customize()"
                     ,"  let current=bufname(\"%\")"
                     ,"  if bufexists(\"customize\")"
                     ,"      set switchbuf+=useopen"
                     ,"      sbuf ^customize$"
                     ,"    else"
                     ,"      new customize"
                     ,"  endif"
                     ,"  set buftype=nofile"
                     ,"  set bufhidden=delete"
                     ,"  set noswapfile"
                     ,"  execute \"%d\""
                     ,"  set tw=0"
                     ,"  map <buffer> _ok :exe getline(3)<cr>:exe getline(4)<cr>:exe getline(5)<cr>:q<cr>"
                     ,"  call append(0,\"let g:haskell_refac_showWindow=\".g:haskell_refac_showWindow)"
                     --,"  call append(0,\"let g:haskell_refac_chasePaths='\".escape(g:haskell_refac_chasePaths,'\\').\"'\")"
                     --,"  call append(0,\"let g:haskell_refac_refactorer='\".escape(g:haskell_refac_refactorer,'\\').\"'\")"
                     ,"  call append(0,\"let g:haskell_refac_chasePaths='\".g:haskell_refac_chasePaths.\"'\")"
                     ,"  call append(0,\"let g:haskell_refac_refactorer='\".g:haskell_refac_refactorer.\"'\")"
                     ,"  call append(0,\"\\\" (you might want to add these lines to your vimrc)\")"
                     ,"  call append(0,\"\\\" To customize, edit values, then type \\\"_ok\\\"\")"
                     ,"endfunction"
                     ,""
                     ,"if has(\"win32\")"
                     ,"  function! Haskell_refac_start()"
                     ,"    let @r=''"
                     ,"    redir @r"
                     ,"    silent execute \":!start /min \".g:haskell_refac_refactorer.\" vim \".escape($VIMRUNTIME,' ').\"\\\\\".v:progname.\" \".v:servername"
                     ,"    au VimLeave * call Haskell_refac_client_cmd('stop')"
                     ,"    redir END"
                     ,"    call Haskell_refac_msg(0,@r)"
                     ,"  "
                     ,"    amenu enable Haskell(Refac).*"
                     ,"    amenu disable Haskell(Refac).start"
                     ,"    amenu disable Haskell(Refac).customize"
                     ,"  endfunction"
                     ,""
                     ,"  function! Haskell_refac_client_cmd(cmd)"
                     ,"    let @r=''"
                     ,"    redir @r"
                     ,"    \"to rescue users of behave mswin"
                     ,"    normal gV"
                     ,"    silent execute \":!start /min \".g:haskell_refac_refactorer_client.\" \".a:cmd"
                     ,"    redir END"
                     ,"    call Haskell_refac_msg(0,@r)"
                     ,"  endfunction"
                     ,"else"
                     ,"  if executable(v:progname)"
                     ,"    function! Haskell_refac_start()"
                     ,"      let @r=''"
                     ,"      redir @r"
                     ,"      silent execute \":!\".g:haskell_refac_refactorer.\" vim \".v:progname.\" \".v:servername.\" >log.txt 2>&1 &\""
                     ,"      au VimLeave * call Haskell_refac_client_cmd('stop')"
                     ,"      redir END"
                     ,"      call Haskell_refac_msg(0,@r)"
                     ,"    "
                     ,"      amenu enable Haskell(Refac).*"
                     ,"      amenu disable Haskell(Refac).start"
                     ,"      amenu disable Haskell(Refac).customize"
                     ,"    endfunction"
                     ,""
                     ,"    function! Haskell_refac_client_cmd(cmd)"
                     ,"      let @r=''"
                     ,"      redir @r"
                     ,"      silent execute \":!\".g:haskell_refac_refactorer_client.\" \".a:cmd"
                     ,"      redir END"
                     ,"      call Haskell_refac_msg(0,@r)"
                     ,"    endfunction"
                     ,"  else"
                     ,"  echoerr \"cannot find path to executable\""
                     ,"  endif"
                     ,"endif"
                     ,""
                     ,"function! Haskell_refac_stop()"
                     ,"  call Haskell_refac_client_cmd('stop')"
                     ,""
                     ,"  amenu disable Haskell(Refac).*"
                     ,"  amenu enable Haskell(Refac).start"
                     ,"  amenu enable Haskell(Refac).version"
                     ,"  amenu enable Haskell(Refac).customize"
                     ,"endfunction"
                     ,""
                     ,"command! RefacStart :call Haskell_refac_start()"
                     ,"command! RefacStop :call Haskell_refac_stop()"
                     ,""
                     ,"function! Haskell_refac_modified(f)"
                     ,"    set switchbuf+=useopen"
                     ,"    if bufexists(a:f)"
                     ,"      execute \":sbuf \".a:f"
                     ,"      edit"
                     ,"    else"
                     ,"      call Haskell_refac_msg(1,\"modified file: \".a:f.\"  (not currently opened)\")"
                     ,"    endif"
                     ,"endfunction"
                     ,""
                     ,"function! Haskell_refac_msg(dialog,msg)"
                     ,"  let visible = bufwinnr('^refac$')"
                     ,"  if bufexists('^refac$')"
                     ,"    set switchbuf+=useopen"
                     ,"    sbuf ^refac$"
                     ,"    call append(line('$'),a:msg)"
                     ,"    if visible==-1 "
                     ,"      hide "
                     ,"    else "
                     ,"      sbuf % "
                     ,"    endif"
                     ,"  else"
                     ,"    hide new refac"
                     ,"    set bufhidden=hide"
                     ,"    set buftype=nofile"
                     ,"    set noswapfile"
                     ,"    call append(line('$'),a:msg)"
                     ,"    hide"
                     ,"  endif"
                     ," if a:dialog"
                     ,"   call confirm(a:msg)"
                     ," endif"
                     ,"endfunction"
                     ,""
                     ]
  selection <- readFile "select.vim"
  putStrLn selection
  where
    gen_function (Entry _ _ c)          = putStrLn $ vim c
    gen_function (Menu _ es)            = mapM_ gen_function es
    gen_command (Entry _ name c)        = putStrLn $ "command! "++capitalise name++" :echo Haskell_refac_"++name++"()"
    gen_command (Menu _ es)             = mapM_ gen_command es
    gen_menu_entry prefix (Menu name es)       = 
                            mapM_ (gen_menu_entry $ prefix++concatMap escape (capitalise name)++".") es
    gen_menu_entry prefix (Entry entry name c) = putStrLn $
                              "\tamenu Haskell(Refac)."
                            ++prefix
                            ++concatMap escape (capitalise entry)++" :"++capitalise name++"<cr>"
                            ++"\n\ttmenu Haskell(Refac)."
                            ++prefix
                            ++concatMap escape (capitalise entry)++description c
    capitalise (h:t) = (toUpper h):t
    escape ' ' = "\\ "
    escape  c  = [c]

gen_emacs_interface = do
  putStrLn $ unlines [";; a minor mode for Haskell refactoring"
                     ,";"
                     ,"; (generated automatically, manual edits may be lost)"
                     ,";"
                     ,";  put this file in your Emacs load-path, and add something "
                     ,";  like this to your .emacs (the hook assumes that you're "
                     ,";  using haskell-mode, and always want to enable the refactoring"
                     ,";  minor mode when entering haskell-mode):"
                     ,";"
                     ,";   (autoload 'haskell-refac-mode \"haskell-refac\""
                     ,";      \"Minor mode for refactoring Haskell programs\" t)"
                     ,";   (add-hook 'haskell-mode-hook 'haskell-refac-mode)"
                     ,";"
                     ,""
                     ,";; menu silently disappears if this defvar is moved after define-minor-mode.."
                     ,"(defvar haskell-refac-mode-map (make-sparse-keymap \"Refactor\"))"
                     ,"(define-minor-mode haskell-refac-mode"
                     ,"  \"Toggle Haskell Refactoring mode."
                     ,"   With no argument, this command toggles the mode. "
                     ,"   Non-null prefix argument turns on the mode."
                     ,"   Null prefix argument turns off the mode."
                     ,""
                     ,"   When Haskell Refactoring mode is enabled, the \\\"Refactor\\\""
                     ,"   menu provides access to a list of refactorings. These are"
                     ,"   implemented as commands of an external refactoring tool"
                     ,"   (default "++refactorer++", see option haskell-refac-refactorer)."
                     ,"   \""
                     ,"  ;; The initial value."
                     ,"  nil"
                     ,"  ;; The indicator for the mode line."
                     ,"  \" "++hare_version++" \""
                     ,"  ;; The minor mode bindings."
                     ,"  nil"
                     ,")"
                     ,""
                     ,";;JP: Quick & dirty: add hooks for the mode we are defining..."
                     ,";; Alternative is not to use define-minor-mode"
                     ,"(add-hook 'haskell-refac-mode-on-hook 'turn-on-haskell-refac)"
                     ,"(add-hook 'haskell-refac-mode-off-hook 'turn-off-haskell-refac)"
                     ,""
                     ,";;JP: moved to top"
                     ,"(defvar haskell-refac-output-marker (make-marker))"
                     ,"(defvar haskell-refac-buffer-list nil)"
                     ,""
                     ,"(defgroup haskell-refac nil \"Haskell refactoring options\")"
                     ,""
                     ,"(defcustom haskell-refac-refactorer \""++refactor_prg++"\""
                     ,"\t\"*Path to Haskell refactoring tool\""
                     ,"\t:type 'file"
                     ,"\t:group 'haskell-refac)"
                     ,""
                     ,"(defcustom haskell-refac-chasePaths '(\".\" "++show preludePath++")"
                     ,"\t\"*List of directories to chase for missing files (including Prelude!)\""
                     ,"\t:type '(repeat directory)"
                     ,"\t:group 'haskell-refac)"
                     ,""
                     ,";;JP: easy-menu should work with Emacs too"
                     ,"(easy-menu-define"
                     ," haskell-refac-menu"
                     ," haskell-refac-mode-map"
                     ," \"Haskell refactorer minor mode menu\""
                     ," '(\"Refactor\""
--                     ,"   [\"New project\" haskell-refac-new   t]"
--                     ,"   [\"Add file\"    haskell-refac-add   t]"
--                     ,"   [\"Chase files\" haskell-refac-chase t]"
--                     ,"   [\"List files\"  haskell-refac-files t]"
--                     ,"   \"-\""
                     ,""
                     ]
  mapM_ gen_menu_entry cmds
  putStrLn $ unlines ["   \"-\""
--                     ,"   [\"Undo\"        haskell-refac-undo  t]     "
--                     ,"   \"-\""
                     ,"   [\"Customize\" ((customize-group 'haskell-refac)) t]"
                     ,"   ))"
                     ,""
                     ,"(defun turn-on-haskell-refac ()"
                     ,"  \"Turn on Haskell Refactoring support\""
                     ,"  (interactive)"
                     ,""
                     ,"  (if (not (buffer-file-name))"
                     ,"    (message-box \"Error: can only refactor if buffer is associated with file\")"
                     ,""
                     ,"    (progn "
                     ,"      ;; keep track of buffers involved in refactoring"
                     ,""
                     ,"      (add-to-list 'haskell-refac-buffer-list (buffer-name) )"
                     ,""
                     ,"      ;; key bindings and menu entries"
                     ,""
                     ,"      (easy-menu-add haskell-refac-menu)"
                     ,"      (hare-selection-init)"
                     ,""
                     ,"      ;; refactoring sub-process, only one, please"
                     ,""
                     ,"      (unless (get-process \"haskell-refac-process\")"
                     ,"        (start-process \"haskell-refac-process\" \"*refac*\""
                     ,"            haskell-refac-refactorer \"emacs\")"
                     ,""
                     ,"        ; does this work on unix as well, or do we need to be os-specific?"
                     ,"        ;;JP: XEmacs doesn't know this"
                     ,"        ;;(set-process-coding-system (get-process \"haskell-refac-process\") "
                     ,"        ;;                           'raw-text-dos 'raw-text-unix)"
                     ,""
                     ,"        (set-marker haskell-refac-output-marker 0 (get-buffer-create \"*refac*\"))"
                     ,""
                     ,"        (set-process-filter (get-process \"haskell-refac-process\") "
                     ,"                            'haskell-refac-process-filter) ) ) )"
                     ,")"
                     ,""

                     ,"(defun haskell-refac-process-filter (proc string)"
                     ,"     (let* ((refac-buffer (process-buffer proc)) "
                     ,"            (refac-end-mark (process-mark proc))"
                     ,"            (refac-buffer-window (display-buffer refac-buffer))"
                     ,"            (modified nil))"
                     ,"       (with-current-buffer refac-buffer"
                     ,"           ;; Insert the text, advancing the process marker."
                     ,"           (goto-char refac-end-mark)"
                     ,"           (insert string)"
                     ,"           (set-marker refac-end-mark (point))"
                     ,"           ;; has the refactorer changed files on disk?"
                     ,"           (save-excursion"
                     ,"               (goto-char (marker-position haskell-refac-output-marker))"
                     ,"               (let ((found nil))"
                     ,"                  (while (setq found (re-search-forward \"^\\\\(message\\\\|modified\\\\): \\\\(.*\\\\)\\n\" nil t)) "
                     ,"                    (progn"
                     ,"                      (let ((cmd (match-string 1)) (param (match-string 2)))"
                     ,"                       (goto-char found)"
                     ,"                       (set-marker haskell-refac-output-marker found)"
                     ,"                       (cond"
                     ,"                         ((string= cmd \"modified\") (setq modified (cons param modified)))"
                     ,"                         ((string= cmd \"message\")  (message-box param))"
                     ,"                         )"
                     ,"                      )"
                     ,"                     ) )"
                     ,"               )"
                     ,"           )"
                     ,"       )"
                     ,"       (save-selected-window "
                     ,"         (select-window refac-buffer-window)"
                     ,"         (goto-char (marker-position refac-end-mark))"
                     ,"         (recenter -2))"
                     ,"       (dolist (f modified) "
                     ,"         (let ((buffer (get-file-buffer f)))"
                     ,"           (if buffer (with-current-buffer buffer (revert-buffer nil t t))"
                     ,"             (message-box (format \"modified unopened file: %s\" f)))))"
                     ,"     )"
                     ,")"

                     ,""
                     ,"(defun turn-off-haskell-refac ()"
                     ,"  \"Turn off Haskell Refactoring support\""
                     ,"  (interactive)"
                     ,"  (delete (buffer-name) haskell-refac-buffer-list)"
                     ,"  (easy-menu-remove haskell-refac-menu)"
                     ,"  (if (null haskell-refac-buffer-list) "
                     ,"    (set-process-filter (get-process \"haskell-refac-process\") nil)"
                     ,"    (kill-process \"haskell-refac-process\")"
                     ,"    (kill-buffer \"*refac*\")"
                     ,"  )"
                     ,"  )"
                     ,""
                     ,";; try to provide command to call pfe with current"
                     ,";; file name and position in file"
                     ,""
                     ,"(defun line-no ()"
                     ,"  \"grmpff. does anyone understand count-lines?\""
                     ,"  (+ (if (eq 0 (current-column)) 1 0)"
                     ,"     (count-lines (point-min) (point)))"
                     ,"  )"
                     ,""
                     ,"(defun line-no-pos (pos)"
                     ,"  \"grmpff. why no parameter to current-column?\""
                     ,"  (save-excursion"
                     ,"    (goto-char pos)"
                     ,"    (+ (if (eq 0 (current-column)) 1 0)"
                     ,"       (count-lines (point-min) (point))))"
                     ,"  )"
                     ,""
                     ,"(defun current-column-pos (pos)"
                     ,"  \"grmpff. why no parameter to current-column?\""
                     ,"  (save-excursion"
                     ,"    (goto-char pos) (current-column))"
                     ,"  )"
                     ]
  mapM_ gen_function cmds
  selection <- readFile "select.el"
  putStrLn selection
  where
    gen_menu_entry (Menu name es)       = do
                      putStrLn $ "(\""++name++"\""
                      mapM_ gen_menu_entry es
                      putStrLn $ ")"
    gen_menu_entry (Entry entry name c) = putStrLn $
-- Xemacs easy menu doesn't support menu tooltips
--                      "   [\""++entry++"\" haskell-refac-"++name++" :help \""++description c++"\"  :active t]"
                      "   [\""++entry++"\" haskell-refac-"++name++"  :active t]"
    gen_function (Entry _ _ c)          = putStrLn $ emacs c
    gen_function (Menu _ es)            = mapM_ gen_function es

help = do
  mapM_ (putStrLn . \(Entry entry name c)->entry++"("++name++")"++description c) cmds
  putStrLn "--"
  putStrLn "help : display this usage information"
  putStrLn "vim  : generate Vim binding"
  putStrLn "emacs: generate Emacs binding"
  putStrLn ""


cmds = [Menu "Projects"
         [editParameters "New project" "new"
            new $ fileNamePar $ 
            comment "Start new project with current file"
         ,editParameters "Add file" "add"
            add $ fileNamePar $ 
            comment "Add current file to project"
         ,editParameters "Chase imports" "chase"
            chase $ optionPar True "chasePaths" $ 
            comment "Chase to add missing files to project"
         ,editParameters "List files" "files"
            files $ 
            comment "List files in project"
        ]

       ,Menu "Names/Scopes"
         [editParameters "Rename" "rename"
            rename $ fileNamePar $ namePar "New name? " $ positionPar $ 
            comment "Rename a variable name"

         ,editParameters "Lift def to top level" "liftToTopLevel"
            liftToTopLevel $ fileNamePar $ positionPar $ 
            comment "Lift a local declaration to top level"
         ,editParameters "Lift def one level" "liftOneLevel"
            liftOneLevel $ fileNamePar $ positionPar $ 
            comment "Lift a local declaration one level up"
         ,editParameters "Demote def" "demote"
            demote $ fileNamePar $ positionPar $ 
            comment "Demote a declaration to where it is used"
         ]

       ,Menu "Definitions"
         [editParameters "Introduce new def" "introNewDef"
            introNewDef $ fileNamePar $ namePar "Name for new definition? " $ regionPar $ 
            comment "Introduce a new definition"
         ,editParameters "Unfold def" "unfoldDef"
            unfoldDef $ fileNamePar $ positionPar $ 
            comment "unfold a definition"
         ,editParameters "Generalise def" "generaliseDef"
            generaliseDef $ fileNamePar $ namePar "name of new parameter? " $ regionPar $ 
            comment "generalise a definition"

         ,editParameters "Remove def" "removeDef"
            removeDef $ fileNamePar $ positionPar $ 
            comment "Remove a definition if it is no used"
         ,editParameters "Duplicate def" "duplicateDef"
            duplicateDef $ fileNamePar $ namePar "Name for duplicate? " $ positionPar $ 
            comment "Duplicate a definition at the same level"
         ,editParameters "Add one parameter" "addOneParameter"
            addOneParameter $ fileNamePar $ namePar "name of new parameter? " $ positionPar $ 
            comment "add parameter (default undefined)"
         ,editParameters "Rm one parameter" "rmOneParameter"
            rmOneParameter $ fileNamePar $ positionPar $ 
            comment "remove unused parameter"
         ,editParameters "Move def to another module" "moveDefBtwMod" 
            moveDefBtwMod $ fileNamePar $ namePar "name of the destination module? " $ positionPar $ 
            comment "move a definition from one module to another module" 

      {-   ,editParameters "Swap Arguments" "swapArgs"
            swapArgs $ fileNamePar $ positionPar $ 
            comment "Swap the first two arguments of a function"
         ,editParameters "From if to case" "ifToCase"
            ifToCase $ fileNamePar $ regionPar $ 
            comment "From if to case"  -}
 
         ]
       
       ,Menu "Import/Export"
         [editParameters "Clean imports" "cleanImports" 
            cleanImports $ fileNamePar $
            comment "tidy up the import list of the current module" 
         ,editParameters "Make import explicit" "mkImpExplicit"
           mkImpExplicit $ fileNamePar $ positionPar $
           comment "make the used entities explicit"
         ,editParameters "Add to export" "addToExport" 
          addToExport $ fileNamePar $ positionPar $
          comment "add an identifier to the export list"
         ,editParameters "Remove from export" "rmFromExport" 
          rmFromExport $ fileNamePar $ positionPar $
          comment "remove an identifier from the export list" 
         ]

       ,Menu "Data types"
         [editParameters "Add field labels" "addFieldLabels" 
          addFieldLabels $ fileNamePar $ positionPar $
          comment "add field labels to a data type declaration"   
         ,editParameters "Add discriminators" "addDiscriminators" 
          addDiscriminators $ fileNamePar $ positionPar $
          comment "add discriminator functions to a data type declaration"   
         ,editParameters "Add constructors" "addConstructors" 
          addConstructors $ fileNamePar $ positionPar $
          comment "add constructor functions to a data type declaration"          
         ,editParameters "eliminate nested patterns" "elimNestedPatterns" 
          elimPatterns  $ fileNamePar $ positionPar $
          comment "eliminate pattern matchings"
         ,editParameters "eliminate patterns" "elimPatterns" 
          elimPatterns  $ fileNamePar $ positionPar $
          comment "eliminate pattern matchings"
         ,editParameters "Create an ADT module" "createADTMod" 
            createADTMod $ fileNamePar $ positionPar $ 
            comment "create an new ADT module"
         ,editParameters "From concrete to abstract data type" "fromAlgebraicToADT" 
            fromAlgebraicToADT $ fileNamePar $ positionPar $ 
            comment "Transform an algebraic data type to an ADT"
         ]
     {-   ,Menu "Duplicate code"
         [editParameters "Detect duplicate code" "detectDuplicates"
          detectDuplicates $
          comment "Detect all the duplicate code in a program"
         ,editParameters "Detect specific duplicate code" "detectSpecDuplicates"
         detectSpecDuplicates $ fileNamePar $ regionPar $
         comment "Detect the duplicates of a specific piece of code"
         ] -}
       ,editParameters "undo" "undo"
          undo $ 
          comment "one step back in refactorer history"
       ]

new :: String -> IO ()
new f = putStrLn $ ">new filename: "++f

add :: String -> IO ()
add f = putStrLn $ ">add filename: "++f

chase :: String -> IO ()
chase d = putStrLn $ ">chase directory: "++d

files :: IO ()
files = putStrLn $ ">files "

undo :: IO ()
undo = putStrLn $ ">undo "

removeDef :: String -> Int -> Int -> IO ()
removeDef f l c = putStrLn $ ">removeDef filename: "++f++" line: "++show l++" column: "++show c

duplicateDef :: String -> String -> Int -> Int -> IO ()
duplicateDef f n l c = putStrLn $ ">duplicateDef filename: "++f++" name for duplicate: "++show n++" line: "++show l++" column: "++show c

liftToTopLevel :: String -> Int -> Int -> IO ()
liftToTopLevel f l c = putStrLn $ ">liftToTopLevel filename: "++f++" line: "++show l++" column: "++show c

liftOneLevel :: String -> Int -> Int -> IO ()
liftOneLevel f l c = putStrLn $ ">liftOneLevel filename: "++f++" line: "++show l++" column: "++show c

demote :: String -> Int -> Int -> IO ()
demote f l c = putStrLn $ ">demote filename: "++f++" line: "++show l++" column: "++show c

rename :: String -> String -> Int -> Int -> IO ()
rename f n l c = putStrLn $ ">rename filename: "++f++" name: "++n++" line: "++show l++" column: "++show c

introNewDef :: String -> String -> Int -> Int -> Int -> Int -> IO ()
introNewDef f n ls cs le ce = putStrLn $ 
    ">introNewDef filename: "++f++" name: "++n
  ++" line: "++show ls++" column: "++show cs
  ++" line: "++show le++" column: "++show ce

unfoldDef :: String -> Int -> Int -> IO ()
unfoldDef f l c = putStrLn $ ">unfoldDef filename: "++f++" line: "++show l++" column: "++show c

addOneParameter :: String -> String -> Int -> Int -> IO ()
addOneParameter f p l c = putStrLn $ ">addOneParameter filename: "++f++" parameter name: "++p++" line: "++show l++" column: "++show c

rmOneParameter :: String -> Int -> Int -> IO ()
rmOneParameter f l c = putStrLn $ ">rmOneParameter filename: "++f++" line: "++show l++" column: "++show c

generaliseDef :: String -> String -> Int -> Int -> Int -> Int -> IO ()
generaliseDef f p ls cs le ce = putStrLn $ 
    ">generaliseDef filename: "++f++" parameter name: "++p
  ++" line: "++show ls++" column: "++show cs
  ++" line: "++show le++" column: "++show ce

moveDefBtwMod :: String -> String -> Int -> Int ->IO ()
moveDefBtwMod f p ls cs = putStrLn $ 
    ">moveDefBtwMod filename: "++f++" module name: "++p
  ++" line: "++show ls++" column: "++show cs

deforest::String->IO()
deforest f = putStrLn $
    "> Eliminate intermidiate list: " ++ f

cleanImports :: String->IO ()
cleanImports f= putStrLn $ 
    ">cleanImports filename: "++f

mkImpExplicit :: String -> Int -> Int -> IO ()
mkImpExplicit f l c = putStrLn $ ">make the imported entities explicit: "++f++" line: "++show l++" column: "++show c

addToExport :: String -> Int -> Int -> IO ()
addToExport f l c = putStrLn $ ">add an identifier to the export list: "++f++" line: "++show l++" column: "++show c

rmFromExport ::String -> Int -> Int -> IO ()
rmFromExport f l c = putStrLn  $ ">remove an identifier from the export list: "++f++" line: "++show l++" column: "++show c 

addFieldLabels ::String -> Int -> Int -> IO ()
addFieldLabels f l c = putStrLn  $ ">add field labels to a data type declaration: "++f++" line: "++show l++" column: "++show c 

addDiscriminators ::String -> Int -> Int -> IO ()
addDiscriminators f l c = putStrLn  $ ">add discriminator functions to a data type declaration: "++f++" line: "++show l++" column: "++show c

addConstructors ::String -> Int -> Int -> IO ()
addConstructors f l c = putStrLn  $ ">add constructor functions to a data type declaration: "++f++" line: "++show l++" column: "++show c

elimPatterns ::String -> Int -> Int -> IO ()
elimPatterns f l c = putStrLn  $ ">eliminate pattern bindings: "++f++" line: "++show l++" column: "++show c

elimNestedPatterns ::String -> Int -> Int -> IO ()
elimNestedPatterns f l c = putStrLn  $ ">eliminate nested pattern bindings: "++f++" line: "++show l++" column: "++show c

createADTMod:: String -> Int -> Int ->IO ()
createADTMod f l c = putStrLn $ 
    ">create an ADT moule: "++f++" line: "++show l++" column: "++show c

fromAlgebraicToADT:: String -> Int -> Int ->IO ()
fromAlgebraicToADT f l c = putStrLn $ 
    ">transform an algebraic data type to an ADT: "++f++" line: "++show l++" column: "++show c


detectDuplicates::IO()
detectDuplicates = putStrLn $
 ">detect duplicate code in a program: "  

detectSpecDuplicates::String->Int->Int->Int->Int->IO()
detectSpecDuplicates f ls cs le ce = putStrLn $ 
    ">detect the duplicates of a specific piece of code: "++f
  ++" line: "++show ls++" column: "++show cs
  ++" line: "++show le++" column: "++show ce
{-
swapArgs:: String -> Int -> Int -> IO ()
swapArgs  f l c = putStrLn $ ">swapArgs filename: "++f++" line: "++show l++" column: "++show c 

ifToCase :: String ->Int -> Int -> Int -> Int -> IO ()
ifToCase f ls cs le ce = putStrLn $
    ">from if to case: "++f
  ++" line: "++show ls++" column: "++show cs
  ++" line: "++show le++" column: "++show ce

-}
---------------------------------------------------------------------------

main = cmdLoop

-- foreign export ccall cmdLoop :: IO ()

cmdLoop :: IO ()
cmdLoop = getArgs >>= doCmd
  {-
  l <- getLine
  doCmd $ words l
  cmdLoop
  -}
  where
    doCmd []         = exitWith ExitSuccess
    doCmd (arg:args) =
        case lookup arg [(entry,c) | (Entry entry name c) <- cmds] of
          Just c  -> cmd c () args
          Nothing | arg=="help"   -> help
          Nothing | arg=="vim"    -> gen_vim_interface 
          Nothing | arg=="emacs"  -> gen_emacs_interface 
          _ -> putStrLn $ "command not found:"++show arg
